/** @file Application.h
 *
 * \author František Toman
 *
 * \version 1.0
 *
 */

/**
 * \class CApplication
 *
 *
 * \brief Jádro aplikace
 *
 * Jedná se o třídu starající se o správu celé aplikace.
 *
 * \author František Toman
 *
 * \version 1.0
 *
 *
 * Kontakt: tomanfra@fit.cvut.cz
 *
 *
 */

#ifndef _CAPP_H_
#define _CAPP_H_

#include "main.h"
#include "GameStateManager.h"

#include "Singleton.h"
#include "Cube.h"

#include "Collection.h"
#include "TextureManager.h"
#include "GameObject3DManager.h"
#include "GameObject2DManager.h"
#include "FontManager.h"
#include "FontPrinter.h"
#include "Renderer.h"

template<class T>
class CApplication: public CSingleton<T> {

public:

	/**
	 * Inicializuje aplikaci
	 * @param argc
	 * @param argv
	 */
	void init(int *argc, char **argv);

	/**
	 * Inicializuje knihovnu DevIL - slouží k načítání a manipulaci obrázků
	 * @return
	 */

	bool initDevIL(void); // inicializace knihovny DevIL

	/**
	 * Inicializace manažerů zdrojů - textury, objekty, písma
	 */
	void initManagers(); // inicializace spravcu

	/**
	 * Spustí smyčku aplikace
	 */
	void run(void);

	/**
	 * Uvolnění zdrojů aplikace
	 */
	static void clear(void);

	/**
	 * Změní scénu hry
	 * @param state - identifikátor scény v manažeru scén
	 */
	void changeState(int state);

	/**
	 * Konstruktor
	 */
	CApplication(void);

	/**
	 * Destruktor
	 */
	~CApplication(void);

	/**
	 * Nastaví titulek okna aplikace
	 * @param str - titulek
	 */
	void setTitle(string str) {
		m_title = str;
	}

	/**
	 * Nastaví titulek okna aplikace
	 * @param str - titulek
	 */
	void setTitle(const char *str) {
		m_title = str;
	}

	/**
	 * Vrátí titulek aplikace
	 * @return
	 */
	string getTitle() {
		return m_title;
	}

	/**
	 * Ukončí aplikaci
	 */
	virtual void terminate();

protected:

	/**
	 * Načte základní zdroje aplikace
	 */
	void loadResources(void); // nacteni zdroju pro hru

	/**
	 * Načte textury
	 */
	virtual void loadTextures(void) = 0; // nacteni textur

	/**
	 * Načte písma
	 */
	virtual void loadFonts(void) = 0; // nacteni fontu

	/**
	 * Vytvoří scény hry
	 */
	virtual void createScenes() = 0; // vytvoreni scen hry


	/**
	 * Zachytí událost pohybu myši
	 * @param x
	 * @param y
	 */
	void mouseMove(int x, int y); // mouse move

	/**
	 * Zachytí událost stisknutí tlačítka myši
	 * @param button
	 * @param state
	 * @param x
	 * @param y
	 */
	void mouseButtons(int button, int state, int x, int y); // mouse button action

	/**
	 * Zachytí událost stisknutí klávesy - ASCII
	 * @param key
	 * @param x
	 * @param y
	 */

	void keyboard(unsigned char key, int x, int y); // keyboard

	/**
	 * Zachytí událost stistknutí klávesy - kurzory, F1, F2, ...
	 * @param key
	 * @param x
	 * @param y
	 */
	void keyboardSpecial(int key, int x, int y);

	/**
	 * Časovač
	 * @param value
	 */
	void timer(int value);

	/**
	 * Vykreslení aktuální scény
	 */
	void renderScene(void); // scene render

	/**
	 * Změna velikosti okna
	 * @param width
	 * @param height
	 */
	void changeSize(int width, int height); // resize

	/**
	 * Vrátí zdali bylo požádáno o ukončení aplikace
	 * @return
	 */

	bool isCallTerminate() {
		return m_terminate;
	}

	/**
	 * Uvolnění zdrojů aplikace
	 */
	void cleanUp();

	static void mouseMoveWrapper(int x, int y);
	static void mouseButtonsWrapper(int button, int state, int x, int y);
	static void keyboardWrapper(unsigned char key, int x, int y);
	static void keyboardSpecialWrapper(int key, int x, int y);
	static void renderSceneWrapper(void);
	static void changeSizeWrapper(int width, int height);
	static void timerWrapper(int value);


	/**
	 * Titulek okna
	 */

	string m_title;

	/**
	 * Ukončení aplikace
	 */

	bool m_terminate;

	/**
	 * Manažer scén
	 */

	CGameStateManager* m_gameStateManager;

};

/**
 *	Konstruktor
 */

template<class T>
CApplication<T>::CApplication(void) {
	m_terminate = false;
	m_gameStateManager = new CGameStateManager();
}

/**
 * Destruktor
 */
template<class T>
CApplication<T>::~CApplication(void) {
	if (m_gameStateManager != NULL) {
		delete m_gameStateManager;
	}
}

/**
 * Inicializace knihoven
 * @param argc
 * @param argv
 */
template<class T>
void CApplication<T>::init(int *argc, char **argv) {

	cout << "* Core initialization" << endl;

	glutInit(argc, argv); // inicializace knihovny GLUT (OpenGL Utility Toolkit)

	/*
	 * inicializace zobrazení:
	 * depth buffer - hloubkový zásobník
	 * double buffering - required to have smooth animation
	 * color model - mód barev - RGBA (červená, zelená, modrá, alfa kanál - průhlednost)
	 *
	 * Podrobnosti: http://www.opengl.org/documentation/specs/glut/spec3/node12.html
	 *
	 */

	glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA | GLUT_STENCIL);
	glutInitWindowSize(800, 600); // nastavení velikosti okna
	glutCreateWindow(getTitle().c_str()); // vytvoříme okno s příslušným titulkem

	glutSetCursor(GLUT_CURSOR_NONE); // schováme kurzor

	// Nastavení callbacků funkcí
	// - vždy ukazatel na statickou metodu dané třídy, která následně volá příslušné nestatické funkce

	glutDisplayFunc(renderSceneWrapper); // vykreslování
	glutReshapeFunc(changeSizeWrapper); // změna velikosti okna
	glutKeyboardFunc(keyboardWrapper); // obsluha klávesnice (ASCII)
	glutSpecialFunc(keyboardSpecialWrapper); // obsluha klávesnice (speciální vsupty - např. šipky, F1, F2, ...)
	glutPassiveMotionFunc(mouseMoveWrapper); // obsluha pohybu myši
	glutMouseFunc(mouseButtonsWrapper); // obsluha stisknutí tlačítek myši
	glutTimerFunc(10, timerWrapper, 0); // časovač

	initDevIL(); // inicializace knihovny DevIL
	loadResources(); // načtení zdrojů
	createScenes(); // vytvoreni scen

	// inicializace kamery
	CCamera* camera = CCamera::getInstance();

	// nastaveni renderu
	CRenderer* renderer = CRenderer::getInstance();
	renderer->blendingEnable();
	renderer->setViewport(800, 600);
	renderer->setCamera(camera);

	// nastaveni vypisovace pisma
	CFontPrinter* fontPrinter = CFontPrinter::getInstance();
	fontPrinter->setFontManager(CFontManager::getInstance());
}

template<class T>
void CApplication<T>::initManagers() {
	m_gameStateManager = new CGameStateManager();
}

/**
 * Inicializace knihovny DevIL - slouží k načítání obrázků z externích zdrojů
 * Podrobnosti: http://openil.sourceforge.net/
 * @return
 */

template<class T>
bool CApplication<T>::initDevIL() {

	// Otestujeme zdali stavíme na správné verzi knihovny DevIL
	// Zdroj: http://openil.sourceforge.net/tuts/tut_step/index.htm
	//

	if (ilGetInteger(IL_VERSION_NUM) < IL_VERSION || iluGetInteger(
			ILU_VERSION_NUM) < ILU_VERSION || ilutGetInteger(ILUT_VERSION_NUM)
			< ILUT_VERSION) {
		cerr << "DevIL version is different...exiting!\n" << endl;
		return false;
	}

	cout << "DevIL initialization ... ";

	// inicializace knihoven
	// il - slouží k načítání a ukládání obrázků
	// ilu - slouží k modifikaci obrázků
	// ilut - propojuje knihovku DevIL s OpenGL

	ilInit();
	iluInit();
	ilutRenderer(ILUT_OPENGL); // podpora OPENGL

	cout << "succeed" << endl;

	return true;
}

/**
 * Spustí hlavní smyčku aplikace
 */
template<class T>
void CApplication<T>::run(void) {
	glutMainLoop(); // hlavní smyčka aplikace
}

// states

template<class T>
void CApplication<T>::changeState(int state) {
	if (m_gameStateManager != NULL) {
		m_gameStateManager->switchStateTo(state);
	}
}

/**
 * Načtení základních zdrojů - písma, textury
 */
template<class T>
void CApplication<T>::loadResources(void) {
	cout << "*Resources loading" << endl;
	cout << "**Textures loading" << endl;
	loadTextures(); // načtení textur
	cout << "**Font loading" << endl;
	loadFonts(); // načtení písem
}

// service methods

/**
 * Časovač
 * @param value
 */
template<class T>
void CApplication<T>::timer(int value) {
	glutPostRedisplay(); // překresli scénu
	glutTimerFunc(1000 / 60, timerWrapper, 1); // zavolej časovač
}

/**
 * Obsluha klávesnice - ASCII
 * @param key
 * @param x
 * @param y
 */
template<class T>
void CApplication<T>::keyboard(unsigned char key, int x, int y) {
	if (m_gameStateManager != NULL) {
		this->m_gameStateManager->getCurrentState()->keyboard(key, x, y);
	}
}

/**
 * Obsluha klávesnice - speciální (F1, F2, kurzory)
 * @param key
 * @param x
 * @param y
 */
template<class T>
void CApplication<T>::keyboardSpecial(int key, int x, int y) {
	if (m_gameStateManager != NULL) {
		this->m_gameStateManager->getCurrentState()->keyboardSpecial(key, x, y);
	}
}

/**
 * Obsluha pohybu myši
 * @param x
 * @param y
 */
template<class T>
void CApplication<T>::mouseMove(int x, int y) {
	if (m_gameStateManager != NULL) {
		this->m_gameStateManager->getCurrentState()->mouseMove(x, y);
	}
}

/**
 * Obsluha tlačítek myši
 * @param button
 * @param state
 * @param x
 * @param y
 */
template<class T>
void CApplication<T>::mouseButtons(int button, int state, int x, int y) {
	if (m_gameStateManager != NULL) {
		this->m_gameStateManager->getCurrentState()->mouseButtons(button,
				state, x, y);
	}
}

/**
 * Vykreslení scény
 */

template<class T>
void CApplication<T>::renderScene(void) {

	if (this->isCallTerminate()) {
		this->cleanUp();
	}

	if (m_gameStateManager != NULL) {
		m_gameStateManager->update();
		m_gameStateManager->render();
	}
}

/**
 * Změma velikosti okna
 * @param width
 * @param height
 */
template<class T>
void CApplication<T>::changeSize(int width, int height) {

	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST); // nejlepsi korekce perspektivy

	if (height == 0) // prevent a divide by zero
		height = 1;

	GLfloat ratio = (GLfloat) width / (GLfloat) height; // screen ratio

	glViewport(0, 0, (GLsizei) width, (GLsizei) height); // set the viewport to be the entire window
	glMatrixMode(GL_PROJECTION); // use matrix projection
	glLoadIdentity(); // reset matrix
	gluPerspective(60, ratio, 0.1, 1000); // set the correct perspective
	glMatrixMode(GL_MODELVIEW); // use model projection

	glEnable(GL_DEPTH_TEST); // hloubkove testovani
	glDepthFunc(GL_LEQUAL); // typ hloubkoveho testovani


}

/*	statické metody	*/

/**
 *	Dealokace alokovaných prostředků
 */
template<class T>
void CApplication<T>::clear(void) {

	cout << "Exit application" << endl;

	if (CSingleton<T>::m_instance != NULL)
		delete CSingleton<T>::m_instance;

	exit(0);

}

/*	statické wrappery	*/

/**
 * Obsluha klávesnice - ASCII
 * @param key
 * @param x
 * @param y
 */
template<class T>
void CApplication<T>::keyboardWrapper(unsigned char key, int x, int y) {

	if (CSingleton<T>::m_instance != NULL) {
		CSingleton<T>::m_instance->keyboard(key, x, y);
	}

}

/**
 * Obsluha klávesnice - speciální
 * @param key
 * @param x
 * @param y
 */

template<class T>
void CApplication<T>::keyboardSpecialWrapper(int key, int x, int y) {
	if (CSingleton<T>::m_instance != NULL) {
		CSingleton<T>::m_instance->keyboardSpecial(key, x, y);
	}
}

/**
 * Obsluha pohybu kurzoru
 * @param x
 * @param y
 */

template<class T>
void CApplication<T>::mouseMoveWrapper(int x, int y) {

	if (CSingleton<T>::m_instance != NULL)
		CSingleton<T>::m_instance->mouseMove(x, y);

}

/**
 * Obsluha tlačítek myši
 * @param button
 * @param state
 * @param x
 * @param y
 */
template<class T>
void CApplication<T>::mouseButtonsWrapper(int button, int state, int x, int y) {

	if (CSingleton<T>::m_instance != NULL)
		CSingleton<T>::m_instance->mouseButtons(button, state, x, y);

}

/**
 * Vykreslení scény
 */
template<class T>
void CApplication<T>::renderSceneWrapper(void) {

	if (CSingleton<T>::m_instance != NULL)
		CSingleton<T>::m_instance->renderScene();

}

/**
 * Změna velikosti okna
 * @param width
 * @param height
 */
template<class T>
void CApplication<T>::changeSizeWrapper(int width, int height) {

	if (CSingleton<T>::m_instance != NULL)
		CSingleton<T>::m_instance->changeSize(width, height);

}

/**
 * Časovač
 * @param value
 */
template<class T>
void CApplication<T>::timerWrapper(int value) {

	if (CSingleton<T>::m_instance != NULL)
		CSingleton<T>::m_instance->timer(value);

}

template<class T>
void CApplication<T>::terminate() {
	this->m_terminate = true;
}

template<class T>
void CApplication<T>::cleanUp() {

	cout << "DISCONNECT GAME STATE MANAGER" << endl;
	if (m_gameStateManager != NULL) {
		CGameStateManager* gm = m_gameStateManager;
		gm->clean();
		m_gameStateManager = NULL;
	}

	cout << "DISCONNECT RENDERS" << endl;
	CRenderer* renderer = CRenderer::getInstance();
	if (renderer != NULL) {
		renderer->disconnect();
		delete renderer;
	}

	cout << "DELETE 2D MANAGER OBJECTS" << endl;
	CGameObject2DManager* object2DManager = CGameObject2DManager::getInstance();
	if (object2DManager != NULL) {
		object2DManager->clean();
		delete object2DManager;
	}

	cout << "DELETE 3D MANAGER OBJECTS" << endl;
	CGameObject3DManager* object3DManager = CGameObject3DManager::getInstance();
	if (object3DManager != NULL) {
		object3DManager->clean();
		delete object3DManager;
	}

	cout << "DELETE TEXTURE MANAGER" << endl;

	CTextureManager* textureManager = CTextureManager::getInstance();
	if (textureManager != NULL)
		;
	{
		textureManager->clean();
		delete textureManager;
	}

	cout << "DELETE FONT MANAGER" << endl;
	CFontManager::deleteInstance();
	CCamera::deleteInstance();

	clear();

}

#endif
